﻿//#define DEBUG
//#define DEBUG_LINE
using System;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using UnityEngine;

namespace JointCorrect
{
    class JointCorrect : MVRScript
    {
        private Atom person;
        
        private List< MorphVariant > allMorphs = new List< MorphVariant >();
        private HashSet< string > allVariants = new HashSet< string >();
        private StorableStringSet enabledVariants;
        private Dictionary< string, JSONStorableBool > variantUIs;

        private MorphVariant[] 
            lFoot_x_p065, lFoot_x_n040,
            rFoot_x_p065, rFoot_x_n040,
            lShin_x_p085, lShin_x_p140,
            rShin_x_p085, rShin_x_p140,
            lThigh_x_p035, lThigh_x_n055, lThigh_x_n115, lThigh_y_p075, lThigh_y_n075, lThigh_z_p085, lThigh_z_n015,
            rThigh_x_p035, rThigh_x_n055, rThigh_x_n115, rThigh_y_n075, rThigh_y_p075, rThigh_z_n085, rThigh_z_p015,

            lHand_z_p080, rHand_z_n080,

            lForearm_y_n100, lForearm_y_n130,
            rForearm_y_p100, rForearm_y_p130,
            lShldr_y_n095, lShldr_y_p040, lShldr_z_n075, lShldr_z_n060, lShldr_z_p035,
            rShldr_y_p095, rShldr_y_n040, rShldr_z_p075, rShldr_z_p060, rShldr_z_n035,
            lCollar_x_n025, lCollar_x_p015, 
            rCollar_x_n025, rCollar_x_p015,
            lCollar_y_n026, lCollar_y_p017, 
            rCollar_y_n017, rCollar_y_p026,
            lCollar_z_n015, lCollar_z_p050, 
            rCollar_z_n050, rCollar_z_p015,

            pelvis_x_p030, pelvis_x_n015, 
            abdomen_x_n020, abdomen_x_p030, 
            abdomen2_x_n020, abdomen2_x_p030, 
            chest_x_p020,
            neck_x_n030, neck_y_p035, neck_y_n035,
            head_x_n045, head_x_p035,
            
            cThigh_x_n055, cThigh_x_n115, cThigh_z_p180, cThigh_z_n030;

        private List< Action > drivers = new List< Action >();
        private List< string > NotFound = new List< string >();

        private StringBuilder rLog = new StringBuilder();
        private StringBuilder mLog = new StringBuilder();
        private JSONStorableString morphOut;
        private JSONStorableString jointOut;
        private JSONStorableBool locked;
        private JSONStorableBool builtin;

        private LineRenderer line;
        
        public override void Init()
        {
            base.Init();

            try
            { 
                person = GetContainingAtom();

                allVariants = new HashSet<string>();
                allVariants.Add( "" );
                allVariants.Add( "gens" );
                enabledVariants = new StorableStringSet( "enabledVariants", VariantsChanged );
                enabledVariants.Add( "" );
                enabledVariants.Add( "gens" );
            
                InitMorphs();
                RegisterString( enabledVariants );
            }
            catch( Exception ex )
            {
                SuperController.LogError( $"JointCorrect did not Init.\n{ex}" );
            }
        }

        public void InitMorphs()
        {
            try
            {
                locked = new JSONStorableBool( "Lock", false );
                builtin = new JSONStorableBool( "Built-in base", true );

                #if DEBUG
                { 
                    var verb = builtin.defaultVal ? "is" : "is NOT";
                    //SuperController.LogMessage( $"JointCorrect {verb} from package." );
                    builtin.defaultVal = false;
                    builtin.val = false;

                    morphOut = new JSONStorableString( "morphout", "" );
                    var dbugField = CreateTextField( morphOut );
                    dbugField.height *= 4;

                    jointOut = new JSONStorableString( "jointout", "" );
                    var joutField = CreateTextField( jointOut, true );
                    joutField.height *= 4;

                    var lockedField = CreateToggle( locked, true );
                }
                #endif

                RegisterBool( builtin );
                CreateToggle( builtin, true );

                #region Collar
                var lCollarBone = GetJoint( "lCollar" ).GetComponent< DAZBone >();
                lCollar_x_n025 = GetMorph( "LCollarX-025" );
                lCollar_x_p015 = GetMorph( "LCollarX+015" );
                lCollar_y_n026 = GetMorph( "LCollarY-026" );
                lCollar_y_p017 = GetMorph( "LCollarY+017" );
                lCollar_z_n015 = GetMorph( "LCollarZ-015" );
                lCollar_z_p050 = GetMorph( "LCollarZ+050" );
                drivers.Add( () => { 
                    Vector3 angles = lCollarBone.GetAnglesDegrees();
                    PrintBone( lCollarBone );

                    SetMorph( lCollar_x_n025, angles.x, 0, -25 );
                    SetMorph( lCollar_x_p015, angles.x, 0, 15 );

                    // Note: Y Inverted
                    SetMorph( lCollar_y_n026, angles.y, 0, 26 );
                    SetMorph( lCollar_y_p017, angles.y, 0, -17 );

                    // Note: Rotation order not honored, Z Inverted
                    SetMorph( lCollar_z_n015, angles.z, 0, 15 );
                    SetMorph( lCollar_z_p050, angles.z, 0, -50 );
                });

                var rCollarBone = GetJoint( "rCollar" ).GetComponent< DAZBone >();
                rCollar_x_n025 = GetMorph( "RCollarX-025" );
                rCollar_x_p015 = GetMorph( "RCollarX+015" );
                rCollar_y_n017 = GetMorph( "RCollarY-017" );
                rCollar_y_p026 = GetMorph( "RCollarY+026" );
                rCollar_z_n050 = GetMorph( "RCollarZ-050" );
                rCollar_z_p015 = GetMorph( "RCollarZ+015" );
                drivers.Add( () => {
                    Vector3 angles = rCollarBone.GetAnglesDegrees();
                    PrintBone( rCollarBone );

                    // Note: Rotation order not honored
                    SetMorph( rCollar_x_n025, angles.x , 0, -25 );
                    SetMorph( rCollar_x_p015, angles.x, 0, 15 );

                    // Note: Y Inverted
                    SetMorph( rCollar_y_n017, angles.y, 0, 17 );
                    SetMorph( rCollar_y_p026, angles.y, 0, -26 );

                    // Note: Rotation order not honored, Z Inverted
                    SetMorph( rCollar_z_n050, angles.z, 0, 50 );
                    SetMorph( rCollar_z_p015, angles.z, 0, -15 );
                });
                #endregion

                #region Foot
                var lFootBone = GetJoint( "lFoot" ).GetComponent< DAZBone >();
                lFoot_x_p065 = GetMorph( "LFootX+065" );
                lFoot_x_n040 = GetMorph( "LFootX-040" );
                drivers.Add( () => { 
                    Vector3 angles = lFootBone.GetAnglesDegrees();
                    PrintBone( lFootBone );

                    SetMorph( lFoot_x_n040, angles.x, 0, -40 );
                    SetMorph( lFoot_x_p065, angles.x, 0, 65 );
                });

                var rFootBone = GetJoint( "rFoot" ).GetComponent< DAZBone >();
                rFoot_x_p065 = GetMorph( "RFootX+065" );
                rFoot_x_n040 = GetMorph( "RFootX-040" );
                drivers.Add( () => {
                    Vector3 angles = rFootBone.GetAnglesDegrees();
                    PrintBone( rFootBone );

                    SetMorph( rFoot_x_n040, angles.x, 0, -40 );
                    SetMorph( rFoot_x_p065, angles.x, 0, 65 );
                });
                #endregion

                #region Forearm
                var lForearmBone = GetJoint( "lForeArm" ).GetComponent< DAZBone >();
                lForearm_y_n100 = GetMorph( "LForearmY-100" );
                lForearm_y_n130 = GetMorph( "LForearmY-130" );
                drivers.Add( () => {
                    Vector3 angles = lForearmBone.GetAnglesDegrees();
                    PrintBone( lForearmBone );

                    // Note: Y Inverted
                    SetMorph( lForearm_y_n100, angles.y, 0, 100 );
                    SetMorph( lForearm_y_n130, angles.y, 100, 130 );
                } );

                var rForearmBone = GetJoint( "rForeArm" ).GetComponent< DAZBone >();
                rForearm_y_p100 = GetMorph( "RForearmY+100" );
                rForearm_y_p130 = GetMorph( "RForearmY+130" );
                drivers.Add( () => { 
                    Vector3 angles = rForearmBone.GetAnglesDegrees();
                    PrintBone( rForearmBone );

                    // Note: Y Inverted
                    SetMorph( rForearm_y_p100, angles.y, 0, -100 );
                    SetMorph( rForearm_y_p130, angles.y, -100, -130 );
                });
                #endregion

                #region Hand
                var lHandBone = GetJoint( "lHand" ).GetComponent< DAZBone >();
                var rHandBone = GetJoint( "rHand" ).GetComponent< DAZBone >();
                lHand_z_p080 = GetMorph( "LHandZ+080" );
                rHand_z_n080 = GetMorph( "RHandZ-080" );
                drivers.Add( () => {
                    Vector3 langles = lHandBone.GetAnglesDegrees();
                    Vector3 rangles = rHandBone.GetAnglesDegrees();
                    PrintBone( lHandBone );
                    PrintBone( rHandBone );

                    SetMorph( lHand_z_p080, langles.z, 0, -80 );
                    SetMorph( rHand_z_n080, rangles.z, 0, 80 );
                } );
                #endregion

                #region Shin

                var lShinBone = GetJoint( "lShin" ).GetComponent< DAZBone >();
                lShin_x_p085 = GetMorph( "LShinX+085" );
                lShin_x_p140 = GetMorph( "LShinX+140" );
                drivers.Add( () => {
                    Vector3 angles = lShinBone.GetAnglesDegrees();
                    PrintBone( lShinBone );

                    SetMorph( lShin_x_p085, angles.x, 0, 85 );
                    SetMorph( lShin_x_p140, angles.x, 85, 140 );
                });

                var rShinBone = GetJoint( "rShin" ).GetComponent< DAZBone >();
                rShin_x_p085 = GetMorph( "RShinX+085" );
                rShin_x_p140 = GetMorph( "RShinX+140" );
                drivers.Add( () => { 
                    Vector3 angles = rShinBone.GetAnglesDegrees();
                    PrintBone( rShinBone );

                    SetMorph( rShin_x_p085, angles.x, 0, 85 );
                    SetMorph( rShin_x_p140, angles.x, 85, 140 );
                });
                #endregion

                #region Shoulder
                var lShldrBone = GetJoint( "lShldr" ).GetComponent< DAZBone >();
                lShldr_z_p035 = GetMorph( "LShldrZ+035" );
                lShldr_z_n060 = GetMorph( "LShldrZ-060" );
                lShldr_z_n075 = GetMorph( "LShldrZ-075" );
                lShldr_y_n095 = GetMorph( "LShldrY-095" );
                lShldr_y_p040 = GetMorph( "LShldrY+040" );
                drivers.Add( () => {
                    Vector3 angles = lShldrBone.GetAnglesDegrees();
                    PrintBone( lShldrBone );

                    // Note: Y Inverted
                    SetMorph( lShldr_y_n095, angles.y, 0, 95 );
                    SetMorph( lShldr_y_p040, angles.y, 0, -40 );

                    // Remove -Z as Y approaches -40
                    var zcor = angles.z * (-NormalizeFloat( angles.y, 0, -40));

                    // Note: DAZ Z = VaM Z Inverted
                    SetMorph( lShldr_z_p035, angles.z, 0, -35 );
                    SetMorph( lShldr_z_n060, angles.z + zcor, 0, 60 );
                    SetMorph( lShldr_z_n075, angles.z + zcor, 60, 75 );
                });

                var rShldrBone = GetJoint( "rShldr" ).GetComponent< DAZBone >();
                rShldr_z_n035 = GetMorph( "RShldrZ-035" );
                rShldr_z_p060 = GetMorph( "RShldrZ+060" );
                rShldr_z_p075 = GetMorph( "RShldrZ+075" );
                rShldr_y_n040 = GetMorph( "RShldrY-040" );
                rShldr_y_p095 = GetMorph( "RShldrY+095" );
                drivers.Add( () => { 
                    Vector3 angles = rShldrBone.GetAnglesDegrees();
                    PrintBone( rShldrBone );

                    // Note: Y probably Inverted
                    SetMorph( rShldr_y_n040, angles.y, 0, 40 );
                    SetMorph( rShldr_y_p095, angles.y, 0, -95 );

                    // Remove +Z as Y approaches 40
                    var zcor = angles.z * (-NormalizeFloat( angles.y, 0, 40 ));

                    // Note: DAZ Z = VaM Z Inverted
                    SetMorph( rShldr_z_n035, angles.z, 0, 35 );
                    SetMorph( rShldr_z_p060, angles.z + zcor, 0, -60 );
                    SetMorph( rShldr_z_p075, angles.z + zcor, -60, -75 );
                });
                #endregion

                #region Thigh
                lThigh_x_p035 = GetMorph( "LThighX+035" );
                lThigh_x_n055 = GetMorph( "LThighX-055" );
                lThigh_x_n115 = GetMorph( "LThighX-115" );
                lThigh_y_p075 = GetMorph( "LThighY+075" );
                lThigh_y_n075 = GetMorph( "LThighY-075" );
                lThigh_z_p085 = GetMorph( "LThighZ+085" );
                lThigh_z_n015 = GetMorph( "LThighZ-015" );

                rThigh_x_p035 = GetMorph( "RThighX+035" );
                rThigh_x_n055 = GetMorph( "RThighX-055" );
                rThigh_x_n115 = GetMorph( "RThighX-115" );
                rThigh_y_n075 = GetMorph( "RThighY-075" );
                rThigh_y_p075 = GetMorph( "RThighY+075" );
                rThigh_z_n085 = GetMorph( "RThighZ-085" );
                rThigh_z_p015 = GetMorph( "RThighZ+015" );

                cThigh_z_p180 = GetMorph( "CThighsZ+180" );
                cThigh_z_n030 = GetMorph( "CThighsZ-030" );
                cThigh_x_n055 = GetMorph( "CThighsX-055" );
                cThigh_x_n115 = GetMorph( "CThighsX-115" );
                var lThighBone = GetJoint( "lThigh" ).GetComponent< DAZBone >();
                var rThighBone = GetJoint( "rThigh" ).GetComponent< DAZBone >();
                drivers.Add( () => { 
                    Vector3 lAngles = lThighBone.GetAnglesDegrees();
                    Vector3 rAngles = rThighBone.GetAnglesDegrees();
                    PrintBone( lThighBone );
                    PrintBone( rThighBone );
                    
                    SetMorph( lThigh_x_n115, lAngles.x, -55, -115 );
                    SetMorph( lThigh_x_n055, lAngles.x, 0, -55 );
                    SetMorph( lThigh_x_p035, lAngles.x, 0, 35 );

                    SetMorph( rThigh_x_n115, rAngles.x, -55, -115 );
                    SetMorph( rThigh_x_n055, rAngles.x, 0, -55 );
                    SetMorph( rThigh_x_p035, rAngles.x, 0, 35 );

                    // Note: DAZ Y = VAM Y Inverted
                    SetMorph( lThigh_y_n075, lAngles.y, 0, 75 );
                    SetMorph( lThigh_y_p075, lAngles.y, 0, -75 );

                    // Note: DAZ Y = VAM Y Inverted
                    SetMorph( rThigh_y_n075, rAngles.y, 0, 75 );
                    SetMorph( rThigh_y_p075, rAngles.y, 0, -75 );

                    
                    // Note: DAZ Z = VAM Z Inverted
                    SetMorph( lThigh_z_n015, lAngles.z, 0, 15 );
                    SetMorph( lThigh_z_p085, lAngles.z, 0, -85 );

                    // Note: DAZ Z = VAM Z Inverted
                    SetMorph( rThigh_z_n085, rAngles.z, 0, 85 );
                    SetMorph( rThigh_z_p015, rAngles.z, 0, -15 );

                    var thighZSeparation = rAngles.z - lAngles.z;
                    PrintRotation( thighZSeparation, "thigh Z separation");
                    SetMorph( cThigh_z_p180, thighZSeparation, 0, 180 );
                    SetMorph( cThigh_z_n030, thighZSeparation, 0, -30 );

                    var thighXCombination = (lAngles.x + rAngles.x) / 2;
                    PrintRotation( thighXCombination, "thigh X combination" );
                    SetMorph( cThigh_x_n055, thighXCombination, 0, -55 );
                    SetMorph( cThigh_x_n115, thighXCombination, -55, -115 );
                });
                #endregion

                #region Trunk
                var pelvisBone = GetJoint( "pelvis" ).GetComponent< DAZBone >();
                pelvis_x_p030 = GetMorph( "TPelvisX+030" );
                pelvis_x_n015 = GetMorph( "TPelvisX-015" );
                drivers.Add( () => { 
                    Vector3 angles = pelvisBone.GetAnglesDegrees();
                    PrintBone( pelvisBone );

                    SetMorph( pelvis_x_n015, angles.x, 0, -15 );
                    SetMorph( pelvis_x_p030, angles.x, 0, 30 );
                });

                var abdomenBone = GetJoint( "abdomen" ).GetComponent< DAZBone >();
                abdomen_x_n020 = GetMorph( "TAbdomenX-020" );
                abdomen_x_p030 = GetMorph( "TAbdomenX+030" );
                drivers.Add( () => {
                    Vector3 angles = abdomenBone.GetAnglesDegrees();
                    PrintBone( abdomenBone );

                    SetMorph( abdomen_x_n020, angles.x, 0, -20 );
                    SetMorph( abdomen_x_p030, angles.x, 0, 30 );
                });

                var abdomen2Bone = GetJoint( "abdomen2" ).GetComponent< DAZBone >();
                abdomen2_x_n020 = GetMorph( "TAbdomen2X-020" );
                abdomen2_x_p030 = GetMorph( "TAbdomen2X+030" );
                drivers.Add( () => { 
                    Vector3 angles = abdomen2Bone.GetAnglesDegrees();
                    PrintBone( abdomen2Bone );

                    SetMorph( abdomen2_x_n020, angles.x, 0, -20 );
                    SetMorph( abdomen2_x_p030, angles.x, 0, 30 );
                });

                var chestBone = GetJoint( "chest" ).GetComponent< DAZBone >();
                chest_x_p020 = GetMorph( "TChestX+020" );
                drivers.Add( () => { 
                    Vector3 angles = chestBone.GetAnglesDegrees();
                    PrintBone( chestBone );

                    SetMorph( chest_x_p020, angles.x, 0, 20 );
                });

                var neckBone = GetJoint( "neck" ).GetComponent< DAZBone >();
                neck_x_n030 = GetMorph( "TNeckX-030" );
                neck_y_p035 = GetMorph( "TNeckY+035" );
                neck_y_n035 = GetMorph( "TNeckY-035" );
                drivers.Add( () => { 
                    Vector3 angles = neckBone.GetAnglesDegrees();
                    PrintBone( neckBone );

                    SetMorph( neck_x_n030, angles.x, 0, -30 );
                    SetMorph( neck_y_p035, angles.y, 0, -35 );
                    SetMorph( neck_y_n035, angles.y, 0, 35 );
                });

                var headBone = GetJoint( "head" ).GetComponent< DAZBone >();
                head_x_n045 = GetMorph( "THeadX-045" );
                head_x_p035 = GetMorph( "THeadX+035" );
                drivers.Add( () => { 
                    Vector3 angles = headBone.GetAnglesDegrees();

                    SetMorph( head_x_n045, angles.x, 0, -45 );
                    SetMorph( head_x_p035, angles.x, 0, 35 );
                });
                #endregion

                #if DEBUG_LINE
                { 
                    line = gameObject.GetComponent< LineRenderer >();
                    if( line == null ) { 
                        line = gameObject.AddComponent< LineRenderer >();
                    }

                    line.positionCount = 5;
                    foreach( Material mat in line.materials )
                    { 
                        line.startColor = Color.blue;
                        line.endColor = Color.blue;
                        line.startWidth = 0.04f;
                        line.endWidth = 0.00f;
                    }
                }
                #endif

                #if DEBUG
                { 
                    var NotFoundArray = NotFound.ToArray();
                    drivers.Add( () => { 
                        if( NotFoundArray.Length > 0 )
                        {
                            mLog.AppendLine( $"Missing morphs:\n  {string.Join( ",\n  ", NotFoundArray )}" );
                        }
                    });
                }
                #endif

                locked.val = false;
            }
            catch ( Exception ex )
            {
                SuperController.LogError( $"JointCorrect did not Init.\n{ex}" );
                locked.val = true;
            }
        }

        public override void InitUI()
        {
            base.InitUI();
            try
            { 
                #region Variants UI
                variantUIs = new Dictionary<string, JSONStorableBool>();
                foreach( var variantName in allVariants )
                {
                    var viewName = $"Variant: {variantName}";
                    if( variantName == "" )
                        viewName = "JointCorrect: Base";
                    else if( variantName == "gens" )
                        viewName = "JointCorrect: Genitalia";

                    var sw = new JSONStorableBool( viewName, enabledVariants.Contains( variantName ), v => ToggleVariant( v, variantName ) );
                    sw.altName = variantName;
                    variantUIs[ variantName ] = sw;
                    CreateToggle( sw, false );
                }
                #endregion
            }
            catch( Exception ex )
            {
                SuperController.LogError( $"JointCorrect did not InitUI.\n{ex}" );
                locked.val = true;
            }
        }

        public void Update()
        {
            if( locked.val )
                return;
            try
            {
                { 
                    mLog.Remove( 0, mLog.Length );
                    rLog.Remove( 0, rLog.Length );
                }
                
                foreach( Action driver in drivers )
                {
                    driver.Invoke();
                }

                #if DEBUG
                { 
                    morphOut.valNoCallback = mLog.ToString();
                    jointOut.valNoCallback = rLog.ToString();
                }
                #endif
            }
            catch( Exception ex )
            {
                locked.val = true;
                SuperController.LogError( "Exception calculating control morphs: " + ex );
            }
        }

        public void OnDisable()
        {
            foreach( MorphVariant m in allMorphs )
               m.morph.morphValue = 0;
        }

        private void VariantsChanged( HashSet< string > enabledVariants )
        {
            foreach( string variant in allVariants )
            {
                JSONStorableBool vbool = null;
                if( variantUIs.TryGetValue( variant, out vbool ) )
                { 
                    vbool.val = enabledVariants.Contains( variant );
                }
            }
        }

        private void ToggleVariant( bool enabled, string variantName )
        {
            if( enabled ) 
                enabledVariants.Add( variantName );
            else
                enabledVariants.Remove( variantName );
        }

        private void DrawAxes( Quaternion rotation, Vector3 position )
        {
            #if ! DEBUG_LINE

            var length = 0.25f;
            
            line.positionCount = 5;
            
            line.startColor = Color.green;
            line.endColor = Color.blue;

            line.SetPosition( 0, position + ((rotation * Vector3.up).normalized * length) );
            line.SetPosition( 1, position );
            line.SetPosition( 2, position + ((rotation * Vector3.right).normalized * length) );
            line.SetPosition( 3, position );
            line.SetPosition( 4, position + ((rotation * Vector3.forward).normalized * length) );
            #endif
        }

        private void SetMorph( MorphVariant[] morphs, float value, float starting, float ending )
        {
            int mask = builtin.val ? BUILTIN : EXTERNAL;
            value = NormalizeFloat( value, starting, ending );
            foreach( MorphVariant m in morphs )
            {
                if( enabledVariants.Contains( m.variant ) && (m.mask & mask) > 0 )
                { 
                    #if DEBUG
                        mLog.AppendLine( $"{m.morph.displayName} <- {value:##0.00}" );
                    #endif
                    m.morph.morphValue = value;
                }
                else
                {
                    m.morph.morphValue = 0;
                }
            }
        }

        private MorphVariant[] GetMorph( string name )
        {
            var mc = (person.GetStorableByID( "geometry" ) as DAZCharacterSelector).morphsControlUI;
            var morphsList = new List< MorphVariant >();
            foreach( DAZMorph morph in mc.GetMorphs() )
            {
                var names = morph.displayName.Split( '.' );
                if( names[0].Equals( name ) )
                {
                    string variant = "";

                    if( names.Length > 1 )
                    {
                        variant = names[1];

                        if( ! allVariants.Contains( variant ) )
                        { 
                            //SuperController.LogMessage( $"Found JCM variant {variant}" );
                            allVariants.Add( variant );
                        }
                    }

                    if( morph.isLatestVersion )
                    { 
                        if( variant == "" || variant == "gens" )
                        {
                            if( morph.isLatestVersion )
                            { 
                                morphsList.Add( new MorphVariant( variant, morph, morph.isInPackage ? BUILTIN : EXTERNAL ) );
                            }
                            else
                            {
                                // Reset any previous-version morphs.
                                morph.morphValue = 0;
                            }
                        }
                        else
                        {
                            morphsList.Add( new MorphVariant( variant, morph, VARIANT ) );
                        }
                    }
                }
            }

            allMorphs.AddRange( morphsList );
            
            if( morphsList.Count == 0 )
            {
                NotFound.Add( name );
            }

            return morphsList.ToArray();
        }

        private Transform GetJoint( string name )
        {
            return person.GetStorableByID( name ).transform;
        }

        private void PrintBone( DAZBone bone )
        {
            #if DEBUG
                Vector3 angles = bone.GetAnglesDegrees();
                rLog.AppendLine( $"{angles.x: 000;-000}, {angles.y: 000;-000}, {angles.z: 000;-000} {bone.name} {bone.rotationOrder}" );
            #endif
        }

        private void PrintRotation( double value, string name )
        {
            #if DEBUG
                rLog.AppendLine( $"{value: 000.00;-000.00} {name}" );
            #endif
        }

        private static float NormalizeFloat( float value, float start, float end )
        {
            return Mathf.Clamp( (value - start) / (end - start), 0, 1 );
        }

        internal const int BUILTIN = 1;
        internal const int EXTERNAL = 2;
        internal const int VARIANT = 3;

        private class MorphVariant
        {
            public int mask;
            public string   variant;
            public DAZMorph morph;

            public MorphVariant( string variant, DAZMorph morph, int mask )
            {
                this.variant = variant;
                this.morph   = morph;
                this.mask = mask;
            }

            public bool IsSame( string[] names )
            {
                if( names[0] == morph.displayName )
                {
                    string var = names.Length > 1 ? names[1] : "";
                    return variant == var;
                }

                return false;
            }
        }

        public class StorableStringSet : JSONStorableString
        {
            private HashSet< string > underlying;
            public delegate void SetStringSetCallback( HashSet< string > value );
            
            public new SetStringSetCallback setCallbackFunction { get; set; }

            public StorableStringSet( string paramName, SetStringSetCallback callback ) : base( paramName, "<proxy>" )
            {
                underlying = new HashSet< string >();
                setCallbackFunction = callback;
            }

            public void Add( string st )
            {
                if( underlying.Add( st ) )
                    setCallbackFunction( underlying );
            }


            public void Remove( string st )
            {
                if( underlying.Remove( st ) )
                    setCallbackFunction( underlying );
            }

            public bool Contains( string st )
            {
                return underlying.Contains( st );
            }

            public override void RestoreFromJSON( JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true )
            {
                underlying.Clear();
                foreach( JSONNode node in jc[name].AsArray )
                {
                    string s = node.Value;
                    underlying.Add( s );
                }

                if( underlying.Count == 0 )
                {
                    underlying.Add( "" );
                    underlying.Add( "gens" );
                }

                setCallbackFunction( underlying );
            }

            public override bool StoreJSON( JSONClass jc, bool includePhysical = true, bool includeAppearance = true, bool forceStore = false )
            {
                var ar = new JSONArray();
                foreach( string s in underlying )
                {
                    ar.Add( s );
                }
                jc[name] = ar;

                return true;
            }
        }
    }
}
